from rnn import *
import torch
from torch import nn

def run_tests():
    test_rnn_cell()
    test_rnn()
    test_attention()

def test_attention():
    print("Testing attention...", end='')
    embed_dim = 256
    num_heads = 1
    dropout = 0.0
    kdim = embed_dim
    vdim = embed_dim

    attention = nn.MultiheadAttention(
        embed_dim,
        num_heads, 
        dropout, 
        bias=True,
        kdim=kdim,
        vdim=vdim, 
    )

    my_attention = SelfAttention(embed_dim, kdim, vdim)
    
    my_attention.query_transform.weight = nn.Parameter(attention.in_proj_weight[:embed_dim])
    my_attention.key_transform.weight = nn.Parameter(attention.in_proj_weight[embed_dim:embed_dim*2])
    my_attention.value_transform.weight = nn.Parameter(attention.in_proj_weight[embed_dim*2:])

    my_attention.query_transform.bias = nn.Parameter(attention.in_proj_bias[:embed_dim])
    my_attention.key_transform.bias = nn.Parameter(attention.in_proj_bias[embed_dim:embed_dim*2])
    my_attention.value_transform.bias = nn.Parameter(attention.in_proj_bias[embed_dim*2:])

    my_attention.output_transform.weight = attention.out_proj.weight
    my_attention.output_transform.bias = attention.out_proj.bias

    seq_len = 10
    sequence = torch.rand(seq_len, embed_dim)

    query = sequence[-1].unsqueeze(0).unsqueeze(0)
    key = sequence.unsqueeze(1)
    value = sequence.unsqueeze(1)

    observe, my_scores = my_attention(sequence)
    hidden_state, scores = attention(query, key, value)
    expect = hidden_state[0, 0]
    expect_scores = scores[0, 0]

    assert torch.allclose(my_scores, expect_scores)
    assert torch.allclose(observe, expect)

    print("Passed")


def test_rnn_cell():
    print("Testing RNNCell...", end='')
    my_cell = RNNCell(256, 128, 'tanh')
    cell = nn.RNNCell(256, 128, bias=True)

    weight_i2h = my_cell.i2h.weight
    bias_i2h = my_cell.i2h.bias
    weight_h2h = my_cell.h2h.weight
    bias_h2h = my_cell.h2h.bias

    cell.weight_ih = weight_i2h
    cell.weight_hh = weight_h2h
    cell.bias_hh = bias_h2h
    cell.bias_ih = bias_i2h

    dummy_input = torch.rand(256)
    dummy_hidden = torch.rand(128)

    my_out = my_cell(dummy_input, dummy_hidden)

    out = cell(dummy_input.unsqueeze(0), dummy_hidden.unsqueeze(0))

    assert torch.allclose(my_out, out)
    
    print("Passed")


def test_rnn():
    print("Testing RNN...", end='')
    input_size = 256
    hidden_size = 128
    rnn = nn.RNN(
        input_size, 
        hidden_size, 
        num_layers=1, 
        bias=True, 
        nonlinearity='tanh', 
        batch_first=True
    )

    my_rnn = RNN(input_size, hidden_size, use_attention=False, nonlinearity='tanh')

    my_rnn.cell.i2h.weight = rnn.weight_ih_l0
    my_rnn.cell.i2h.bias = rnn.bias_ih_l0
    my_rnn.cell.h2h.weight = rnn.weight_hh_l0
    my_rnn.cell.h2h.bias = rnn.bias_hh_l0

    h0 = torch.zeros(hidden_size)
    input = torch.rand(10, input_size)

    my_hidden, my_output = my_rnn(input)

    hidden, output = rnn(input.unsqueeze(0), h0.unsqueeze(0).unsqueeze(0))

    hidden = hidden[0]
    output = output[0, 0]
    assert torch.allclose(my_hidden, hidden, atol=1e-5)
    assert torch.allclose(my_output[-1], output, atol=1e-5)
    
    print("Passed")


run_tests()